@@ -61,6 +61,11 @@ typedef struct {
 	TrackerResource *resource;
 } UpdateResource;
 
+typedef struct {
+	TrackerSparqlStatement *stmt;
+	GHashTable *parameters;
+} UpdateStatement;
+
 typedef struct {
 	gchar *query;
 	TrackerRdfFormat format;
@@ -88,6 +93,7 @@ typedef enum {
 	TASK_TYPE_UPDATE_BLANK,
 	TASK_TYPE_UPDATE_RESOURCE,
 	TASK_TYPE_UPDATE_BATCH,
+	TASK_TYPE_UPDATE_STATEMENT,
 	TASK_TYPE_RELEASE_MEMORY,
 	TASK_TYPE_SERIALIZE,
 	TASK_TYPE_DESERIALIZE,
@@ -172,17 +178,24 @@ update_resource (TrackerData      *data,
                  GError          **error)
 {
 	GError *inner_error = NULL;
+	GHashTable *visited;
 
 	tracker_data_begin_transaction (data, &inner_error);
 	if (inner_error)
 		goto error;
 
+	visited = g_hash_table_new_full (NULL, NULL, NULL,
+	                                 (GDestroyNotify) tracker_rowid_free);
+
 	tracker_data_update_resource (data,
 	                              graph,
 	                              resource,
 	                              NULL,
+	                              visited,
 	                              &inner_error);
 
+	g_hash_table_unref (visited);
+
 	if (inner_error) {
 		tracker_data_rollback_transaction (data);
 		goto error;
@@ -207,6 +220,8 @@ convert_format (TrackerRdfFormat format)
 		return TRACKER_SERIALIZER_FORMAT_TTL;
 	case TRACKER_RDF_FORMAT_TRIG:
 		return TRACKER_SERIALIZER_FORMAT_TRIG;
+	case TRACKER_RDF_FORMAT_JSON_LD:
+		return TRACKER_SERIALIZER_FORMAT_JSON_LD;
 	default:
 		g_assert_not_reached ();
 	}
@@ -276,6 +291,15 @@ update_thread_func (gpointer data,
 	case TASK_TYPE_UPDATE_BATCH:
 		tracker_direct_batch_update (task_data->data, priv->data_manager, &error);
 		break;
+	case TASK_TYPE_UPDATE_STATEMENT: {
+		UpdateStatement *data = task_data->data;
+
+		tracker_direct_statement_execute_update (data->stmt,
+		                                         data->parameters,
+		                                         NULL,
+		                                         &error);
+		break;
+	}
 	case TASK_TYPE_RELEASE_MEMORY:
 		tracker_data_manager_release_memory (priv->data_manager);
 		update_timestamp = FALSE;
@@ -568,9 +592,7 @@ get_event_cache_ht (TrackerNotifier *notifier)
 
 	events = g_object_get_qdata (G_OBJECT (notifier), tracker_direct_notifier_quark ());
 	if (!events) {
-		events = g_hash_table_new_full (tracker_rowid_hash,
-		                                tracker_rowid_equal,
-		                                (GDestroyNotify) tracker_rowid_free,
+		events = g_hash_table_new_full (g_str_hash, g_str_equal, NULL,
 		                                (GDestroyNotify) _tracker_notifier_event_cache_free);
 		g_object_set_qdata_full (G_OBJECT (notifier), tracker_direct_notifier_quark (),
 		                         events, (GDestroyNotify) g_hash_table_unref);
@@ -581,18 +603,22 @@ get_event_cache_ht (TrackerNotifier *notifier)
 
 static TrackerNotifierEventCache *
 lookup_event_cache (TrackerNotifier *notifier,
-                    TrackerRowid     graph_id,
                     const gchar     *graph)
 {
 	TrackerNotifierEventCache *cache;
 	GHashTable *events;
 
+	if (!graph)
+		graph = "";
+
 	events = get_event_cache_ht (notifier);
-	cache = g_hash_table_lookup (events, &graph_id);
+	cache = g_hash_table_lookup (events, graph);
 
 	if (!cache) {
 		cache = _tracker_notifier_event_cache_new (notifier, graph);
-		g_hash_table_insert (events, tracker_rowid_copy (&graph_id), cache);
+		g_hash_table_insert (events,
+		                     (gpointer) tracker_notifier_event_cache_get_graph (cache),
+		                     cache);
 	}
 
 	return cache;
@@ -602,13 +628,12 @@ lookup_event_cache (TrackerNotifier *notifier,
  * (always the same one though), handle with care.
  */
 static void
-insert_statement_cb (TrackerRowid  graph_id,
-                     const gchar  *graph,
+insert_statement_cb (const gchar  *graph,
                      TrackerRowid  subject_id,
                      TrackerRowid  predicate_id,
                      TrackerRowid  object_id,
-                     GPtrArray   *rdf_types,
-                     gpointer     user_data)
+                     GPtrArray    *rdf_types,
+                     gpointer      user_data)
 {
 	TrackerNotifier *notifier = user_data;
 	TrackerSparqlConnection *conn = _tracker_notifier_get_connection (notifier);
@@ -620,7 +645,7 @@ insert_statement_cb (TrackerRowid  graph_id,
 	TrackerClass *new_class = NULL;
 	guint i;
 
-	cache = lookup_event_cache (notifier, graph_id, graph);
+	cache = lookup_event_cache (notifier, graph);
 
 	if (predicate_id == tracker_property_get_id (rdf_type)) {
 		const gchar *uri;
@@ -646,13 +671,12 @@ insert_statement_cb (TrackerRowid  graph_id,
 }
 
 static void
-delete_statement_cb (TrackerRowid  graph_id,
-                     const gchar  *graph,
+delete_statement_cb (const gchar  *graph,
                      TrackerRowid  subject_id,
                      TrackerRowid  predicate_id,
                      TrackerRowid  object_id,
-                     GPtrArray   *rdf_types,
-                     gpointer     user_data)
+                     GPtrArray    *rdf_types,
+                     gpointer      user_data)
 {
 	TrackerNotifier *notifier = user_data;
 	TrackerSparqlConnection *conn = _tracker_notifier_get_connection (notifier);
@@ -664,7 +688,7 @@ delete_statement_cb (TrackerRowid  graph_id,
 	TrackerClass *class_being_removed = NULL;
 	guint i;
 
-	cache = lookup_event_cache (notifier, graph_id, graph);
+	cache = lookup_event_cache (notifier, graph);
 
 	if (predicate_id == tracker_property_get_id (rdf_type)) {
 		const gchar *uri;
@@ -893,13 +917,22 @@ tracker_direct_connection_query_finish (TrackerSparqlConnection  *self,
 
 static TrackerSparqlStatement *
 tracker_direct_connection_query_statement (TrackerSparqlConnection  *self,
-                                           const gchar              *query,
-                                           GCancellable             *cancellable,
-                                           GError                  **error)
+                                            const gchar              *query,
+                                            GCancellable             *cancellable,
+                                            GError                  **error)
 {
 	return TRACKER_SPARQL_STATEMENT (tracker_direct_statement_new (self, query, error));
 }
 
+static TrackerSparqlStatement *
+tracker_direct_connection_update_statement (TrackerSparqlConnection  *self,
+                                            const gchar              *query,
+                                            GCancellable             *cancellable,
+                                            GError                  **error)
+{
+	return TRACKER_SPARQL_STATEMENT (tracker_direct_statement_new_update (self, query, error));
+}
+
 static void
 tracker_direct_connection_update (TrackerSparqlConnection  *self,
                                   const gchar              *sparql,
@@ -1494,6 +1527,7 @@ tracker_direct_connection_class_init (TrackerDirectConnectionClass *klass)
 	sparql_connection_class->query_async = tracker_direct_connection_query_async;
 	sparql_connection_class->query_finish = tracker_direct_connection_query_finish;
 	sparql_connection_class->query_statement = tracker_direct_connection_query_statement;
+	sparql_connection_class->update_statement = tracker_direct_connection_update_statement;
 	sparql_connection_class->update = tracker_direct_connection_update;
 	sparql_connection_class->update_async = tracker_direct_connection_update_async;
 	sparql_connection_class->update_finish = tracker_direct_connection_update_finish;
@@ -1544,21 +1578,50 @@ tracker_direct_connection_class_init (TrackerDirectConnectionClass *klass)
 	g_object_class_install_properties (object_class, N_PROPS, props);
 }
 
-TrackerDirectConnection *
+TrackerSparqlConnection *
 tracker_direct_connection_new (TrackerSparqlConnectionFlags   flags,
                                GFile                         *store,
                                GFile                         *ontology,
                                GError                       **error)
 {
-	g_return_val_if_fail (!store || G_IS_FILE (store), NULL);
-	g_return_val_if_fail (!ontology || G_IS_FILE (ontology), NULL);
-	g_return_val_if_fail (!error || !*error, NULL);
+	return g_initable_new (TRACKER_TYPE_DIRECT_CONNECTION,
+	                       NULL, error,
+	                       "flags", flags,
+	                       "store-location", store,
+	                       "ontology-location", ontology,
+	                       NULL);
+}
 
-	return g_object_new (TRACKER_TYPE_DIRECT_CONNECTION,
-	                     "flags", flags,
-	                     "store-location", store,
-	                     "ontology-location", ontology,
-	                     NULL);
+void
+tracker_direct_connection_new_async (TrackerSparqlConnectionFlags  flags,
+                                     GFile                        *store,
+                                     GFile                        *ontology,
+                                     GCancellable                 *cancellable,
+                                     GAsyncReadyCallback           cb,
+                                     gpointer                      user_data)
+{
+	g_async_initable_new_async (TRACKER_TYPE_DIRECT_CONNECTION,
+	                            G_PRIORITY_DEFAULT,
+	                            cancellable,
+	                            cb,
+	                            user_data,
+	                            "flags", flags,
+	                            "store-location", store,
+	                            "ontology-location", ontology,
+	                            NULL);
+}
+
+TrackerSparqlConnection *
+tracker_direct_connection_new_finish (GAsyncResult  *res,
+                                      GError       **error)
+{
+	GAsyncInitable *initable;
+
+	initable = g_task_get_source_object (G_TASK (res));
+
+	return TRACKER_SPARQL_CONNECTION (g_async_initable_new_finish (initable,
+	                                                               res,
+	                                                               error));
 }
 
 TrackerDataManager *
@@ -1640,3 +1703,87 @@ tracker_direct_connection_update_batch_finish (TrackerDirectConnection  *conn,
 
 	return TRUE;
 }
+
+static UpdateStatement *
+update_statement_data_new (TrackerSparqlStatement *stmt,
+                           GHashTable             *parameters)
+{
+	UpdateStatement *data;
+
+	data = g_new0 (UpdateStatement, 1);
+	data->stmt = g_object_ref (stmt);
+	data->parameters = parameters ? g_hash_table_ref (parameters) : NULL;
+
+	return data;
+}
+
+static void
+update_statement_data_free (UpdateStatement *data)
+{
+	g_object_unref (data->stmt);
+	g_clear_pointer (&data->parameters, g_hash_table_unref);
+	g_free (data);
+}
+
+gboolean
+tracker_direct_connection_execute_update_statement (TrackerDirectConnection  *conn,
+                                                    TrackerSparqlStatement   *stmt,
+                                                    GHashTable               *parameters,
+                                                    GError                  **error)
+{
+	TrackerDirectConnectionPrivate *priv;
+	GError *inner_error = NULL;
+
+	priv = tracker_direct_connection_get_instance_private (conn);
+
+	g_mutex_lock (&priv->mutex);
+	tracker_direct_statement_execute_update (stmt, parameters, NULL, &inner_error);
+	tracker_direct_connection_update_timestamp (conn);
+	g_mutex_unlock (&priv->mutex);
+
+	if (inner_error) {
+		g_propagate_error (error, inner_error);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+void
+tracker_direct_connection_execute_update_statement_async (TrackerDirectConnection  *conn,
+                                                          TrackerSparqlStatement   *stmt,
+                                                          GHashTable               *parameters,
+                                                          GCancellable             *cancellable,
+                                                          GAsyncReadyCallback       callback,
+                                                          gpointer                  user_data)
+{
+	TrackerDirectConnectionPrivate *priv;
+	GTask *task;
+
+	priv = tracker_direct_connection_get_instance_private (conn);
+
+	task = g_task_new (stmt, cancellable, callback, user_data);
+	g_task_set_task_data (task,
+	                      task_data_query_new (TASK_TYPE_UPDATE_STATEMENT,
+	                                           update_statement_data_new (stmt, parameters),
+	                                           (GDestroyNotify) update_statement_data_free),
+	                      (GDestroyNotify) task_data_free);
+
+	g_thread_pool_push (priv->update_thread, task, NULL);
+}
+
+gboolean
+tracker_direct_connection_execute_update_statement_finish (TrackerDirectConnection  *conn,
+                                                           GAsyncResult             *res,
+                                                           GError                  **error)
+{
+	GError *inner_error = NULL;
+
+	g_task_propagate_boolean (G_TASK (res), &inner_error);
+	if (inner_error) {
+		g_propagate_error (error, _translate_internal_error (inner_error));
+		return FALSE;
+	}
+
+	return TRUE;
+}
